"""A general purpose module that includes simplified pickling/unpickling
functions for saving arbitrarily complex python objects, and functions for
performing multi-threaded calculations."""

import sys
import copy

try:
    unicode = unicode
except NameError:
    unicode = str


class PyARTSError(Exception):
    pass


class PyARTSWarning(UserWarning):
    pass


def quotify(s):
    """Adds quotation marks around the string, if not present.

    This is necessary for string literals in ARTS.
    Does not take care of escaping anything.

    >>> print(quotify("Hello, world!"))
    "Hello, world!"
    """
    if s.startswith('"') and s.endswith('"'):
        return s
    else:
        return '"%s"' % s


def dict_combine_with_default(in_dict, default_dict):
    """A useful function for dealing with dictionary function input.  Combines
    parameters from in_dict with those from default_dict with the output
    having default_dict values for keys not present in in_dict"""
    out_dict = copy.deepcopy(in_dict)
    for key in default_dict.keys():
        out_dict[key] = copy.deepcopy(in_dict.get(key, default_dict[key]))
    return out_dict


def convert_to_bytes(s):
    if isinstance(s, unicode):
        return s.encode('UTF-8')
    else:
        return s


def convert_to_string(s):
    if isinstance(s, (bytes)):
        return s.decode('UTF-8')
    else:
        return s


def force_encoded_string_output(func):
    """ Decorator for __repr__
    """

    if sys.version_info.major < 3:

        def _func(*args, **kwargs):
            return func(*args, **kwargs).encode(sys.stdout.encoding or 'utf-8')

        return _func

    else:
        return func



"""
Create, load and save datafiles in the Arts XML format.

For XML output, the main class is XMLfile.  An XMLfile object is
initialised with a filename or a stream.

>>> from PyARTS import artsXML
>>> testfile=artsXML.XMLfile('a_test_file.xml')

Arts data objects are then added to the file with the add method.

>>> a_tensor=ones([3,5,6],np.float32)
>>> testfile.add(a_tensor)

Context managers are also supported:

>>> with artsXML.XMLfile("a_test_file.xml") as tf:
...     tf.add(my_tensor)

Then you don't need to worry about closing the file, as the context
manager takes care of this.

The ARTS data type to save is is determined automatically from the python
type. This mapping is specified by the dictionary artsXML.MAPPING. For
instance, [[1,2,3],[4,5,6]] would be saved as an ArrayOfArrayOfIndex, whereas
[array([1,2,3]),array([4,5,6])] would be saved as an ArrayOfVector. Note that
it is not guaranteed that ARTS actually understands this! For Tensor type
objects, the tag name (eg. Tensor3) and the size attributes are determined
automatically by the shape of the np array.

The file must then be closed with the close() method:

>>> testfile.close()

A shortcut save function is available.  Using save the above
is achieved in one line.

>>> artsXML.save(a_tensor,'a_test_file.xml')

Some more complicated structures, like SingleScatteringData objects, have
their own save methods which utilize this module.

The load function is a low-level function that returns a dictionary structure
reflecting the structure of the XML file. If you already know the type of the
data you are going to load, you can use the particular constructor, such as
SingleScatteringData.load or ArrayOfLatLonGriddedField3.load. All of these
will use artsXML.load as a backend.
"""

from __future__ import print_function

import numpy as np
import copy

from io import StringIO

from io import BytesIO

try:
    unicode = unicode
except NameError:
    # 'unicode' is undefined, must be Python 3
    unicode = str
    basestring = (str, bytes)

import contextlib
import gzip

from . import general

import xml.sax
import xml.sax.saxutils

from collections import OrderedDict

##CONSTANTS

ARTS_DIM_LABELS = ["ncols", "nrows", "npages", "nbooks",
                   "nshelves", "nvitrines", "nlibraries"]

ARTS_TENSOR_NAMES = ["Vector", "Matrix", "Tensor3", "Tensor4",
                     "Tensor5", "Tensor6", "Tensor7"]

ARTS_TEXT_NAMES = ["String", "comment", "SpeciesTag"]

ARTS_SPARSE_NAMES = ["RowIndex", "ColIndex", "SparseData"]

# controls the verbosity of all classes and functions in this module
VERBOSITY = False


##CLASSES

class Sparse(object):
    """Sparse object

    Attributes attributes are gotten via XML parsing
    (nrows, ncols, etc.)
    """

    def __init__(self, attributes):
        if VERBOSITY:
            print("creating Sparse matrix")
        dimlist = []
        dimlist.append(int(attributes.getValue(ARTS_DIM_LABELS[1])))
        dimlist.append(int(attributes.getValue(ARTS_DIM_LABELS[0])))
        if VERBOSITY:
            print("dimensions:")
            print(dimlist)
        self.dimlist = dimlist
        self.sparse = {}

    def setdata(self, tag, v):
        if tag in ARTS_SPARSE_NAMES:
            self.sparse[tag] = v
        else:
            raise TypeError("Unknown member in Sparse type: %s" % tag)

    def csc_matrix(self):
        """ Returns a scipy sparse object """
        import scipy.sparse

        return scipy.sparse.csc_matrix(
            (self.sparse['SparseData'],
             (self.sparse['RowIndex'], self.sparse['ColIndex'])),
            self.dimlist)


class Tensor(object):
    """Tensor object

    Rank, attributes attributes are gotten via XML parsing
    (nelem, nlibraries, etc.)
    """

    def __init__(self, rank, attributes):
        if VERBOSITY:
            print("creating Tensor of rank {0}".format(rank))
        dimlist = []
        if rank == 1:
            dimlist.append(int(attributes.getValue("nelem")))
        else:
            for i in range(rank):
                dimlist.append(int(attributes.getValue(ARTS_DIM_LABELS[i])))
        # put the highest dimensions first
        dimlist.reverse()
        if VERBOSITY:
            print("dimensions:")
            print(dimlist)
        # fill with nans
        Z = np.zeros(dimlist)
        Z.fill(np.nan)
        self.array = Z
        self.dimlist = dimlist

    def setdata(self, v):
        if v.size != np.prod(self.dimlist):
            raise TypeError("Shape mismatch: data %s, dimlist %s"
                            % (v.shape, self.dimlist))
        self.array = v
        self.fixshape()

    def fixshape(self):
        if VERBOSITY:
            print("raw data shape = ")
            print(self.array.shape)
        self.array.shape = self.dimlist


# Mapping between Python and ARTS types
MAPPING = {
    list: 'Array',
    int: 'Index',
    float: 'Numeric',
    str: 'String',
    np.ndarray: 'Tensor',
    Sparse: 'Sparse',
}


class XML_Obj(object):
    def __init__(self, tag, attributes=None):
        if attributes is None:
            attributes = {}

        self.tag = tag
        self.s = BytesIO()
        self.write(
            (u'<' + tag + _attributeconvert(attributes) + '>\n'))

    def write(self, content):
        self.s.write(general.convert_to_bytes(content))
        return self

    def finalise(self):
        self.write(('</' + self.tag + '>\n'))
        self.str = self.s.getvalue()
        self.s.close()
        return self


class XMLfile(object):
    """Arts XML output class.

    Initialise with a filename or object.
    Warning: this will open the file for writing, thus over-writing any
    existing data!
    """

    def __init__(self, f, header=True):
        if isinstance(f, basestring):
            self.filename = f
            self.file = open(f, 'wb')
        else:
            self.file = f
        if header:
            self.file.write('<?xml version="1.0"?>\n'.encode())
            self.file.write('<arts format="ascii" version="1">\n'.encode())
        self.header = header

    def _addnumber(self, tag, number, attributes={}):
        self.file.write(
            general.convert_to_bytes(number_to_xml(tag, number, attributes)))

    def _addText(self, tag, text, attributes={}):
        self.file.write(
            general.convert_to_bytes(text_to_xml(tag, text, attributes)))

    def _addTensor(self, tensor):
        """
        This method takes a np array argument and stores it in
        the arts XML format with the appropriate Tag
        (e.g. Vector, Matrix, ...)
        """
        tensor_to_xml(tensor, self.file)

    def add(self, data):
        """
        Determine the ARTS type from the python type:
        Python list => ARTS array,
        np.ndarray => ARTS Vector/Tensor,
        Python string => ARTS string,
        Python integer => ARTS Index,
        Python float => ARTS Numeric
        """
        if 'np.int' in str(type(data)):
            self._addnumber('Index', data)

        # Python list => ARTS array
        elif MAPPING[type(data)] == 'Array':
            self.file.write(('<Array type="%s" nelem="%d">\n'
                             % (get_arts_type(data[0]), len(data))).encode())
            for x in data:
                self.add(x)
            self.file.write('</Array>\n'.encode())
        elif MAPPING[type(data)] == 'Tensor':
            self._addTensor(data)
        elif MAPPING[type(data)] == 'String':
            self._addString(data)
        elif MAPPING[type(data)] in ['Numeric', 'Index']:
            self._addnumber(MAPPING[type(data)], data)

    def _addString(self, data):
        """adds a String"""
        self.file.write('<String>"'.encode())
        self.file.write(xml.sax.saxutils.escape(data).encode())
        self.file.write('"</String>\n'.encode())

    def close(self, really=True):
        """This must be called to finalise the XML file"""
        if self.header:
            self.file.write(general.convert_to_bytes('</arts>\n'))
        if really:
            self.file.close()

    def write(self, s):
        self.file.write(general.convert_to_bytes(s))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class Handler(xml.sax.ContentHandler):
    """
    XML handler for Arts files.

    This is much more general then the old one. Unlike the previous handler,
    this handler works by continutally updating a list of strings that
    constitute a reference for the eventual location of an object within
    the overall data structure. With the exception of Arrays, which are
    represented by lists, every XML tag becomes a dictionary key. In cases
    where the tag line has a name attribute, the dictionary key becomes the
    name string. This behaviour is enabled by default but can be controlled
    with the une_names argument on initialisation.

    Normally it should not be necessary to use this function directly, instead
    use the newLoad function below.
    """

    def __init__(self, use_names=True):
        """
        If use_names is True, the handler will use any present "name"
        attributes as keys for each object. Otherwise tags are used
        """
        self.CurrentHierarchy = []
        self.datastruct = OrderedDict()
        self.reference_str_list = ["self.datastruct"]
        # this is used to ensure nothing is overwritten
        self.assignment_str_list = []
        self.use_names = use_names

    def characters(self, text):
        try:
            self.buff.write(text)
        except ValueError:
            pass  # for historical reasons

    def startDocument(self):
        if VERBOSITY:
            print("Starting to parse XML")

    def startElement(self, tag, attributes):
        """Called when a new element starts"""

        def advance_name_if_nec(name, reference_str_list):
            if (eval('"' + name + '"'
                         + " in " + "".join(self.reference_str_list))):
                # then change the name
                if VERBOSITY:
                    print('already have ' + name)
                if (len(name.split()) == 1):
                    name = name + ' 0'
                else:
                    label = name.split()[0]
                    number = int(name.split()[1])
                    name = label + ' ' + str(number + 1)
                if VERBOSITY:
                    print("name changed to " + name)
                name = advance_name_if_nec(name, reference_str_list)
            return name

        self.buff = StringIO()
        if self.use_names:
            name = attributes.get("name", tag)
        else:
            name = tag
        if VERBOSITY:
            print("begin Name: " + name + "  Tag: " + tag)
        # Use tag to modify reference string for subsequent data
        if tag == "Array":
            name = advance_name_if_nec(name, self.reference_str_list)
            self.CurrentHierarchy.append(name)

            # [{}]*n gives n references to the same dict!!!!
            list_of_empty_dicts = []
            for i in range(int(attributes["nelem"])):
                list_of_empty_dicts.append(OrderedDict())
            eval("".join(self.reference_str_list)
                 + '.update({"' + name + '":list_of_empty_dicts})'
                 )
            self.reference_str_list.append('["' + name + '"]')
            self.reference_str_list.append("[0]")
            self.CurrentHierarchy.append(0)
        else:
            # check that new object doesn't already exist

            name = advance_name_if_nec(name, self.reference_str_list)

            eval("".join(self.reference_str_list)
                 + '.update({"' + name + '":OrderedDict()})')
            self.reference_str_list.append('["' + name + '"]')
            self.CurrentHierarchy.append(name)

            # If appropriate, create a tensor object
        if tag in ARTS_TENSOR_NAMES:
            rank = ARTS_TENSOR_NAMES.index(tag) + 1
            self.CurrentObject = Tensor(rank, attributes)
        elif tag == 'Sparse':
            self.CurrentObject = Sparse(attributes)

    def text2obj(self, tag, text):
        """Converts XML-text to appropiate object"""

        text = text.strip()
        if tag in ARTS_TENSOR_NAMES:
            self.CurrentObject.setdata(np.loadtxt(BytesIO(text.encode())))
            # self.CurrentObject.fixshape()
            return self.CurrentObject.array
        elif tag in ARTS_SPARSE_NAMES:
            if tag[3:] == 'Index':
                self.CurrentObject.setdata(tag, np.loadtxt(StringIO(text),
                                                           dtype=np.int32))
            else:
                self.CurrentObject.setdata(tag, np.loadtxt(StringIO(text)))
            return self.CurrentObject.sparse[tag]
        elif tag in ARTS_TEXT_NAMES:
            return text
        elif tag == "Index":
            return np.int64(text)
        elif tag == "Numeric":
            return np.float64(text)
        elif tag == "Sparse":
            return self.CurrentObject
        else:
            if text.strip():
                raise TypeError("Cannot convert to obj: %s" % tag)
            else:
                return None

    def endElement(self, tag):
        if not self.buff.closed:
            text = self.buff.getvalue()
            self.buff.close()
        else:
            text = ""
        if VERBOSITY:
            print("end  Tag: " + tag)
        obj = self.text2obj(tag, text)
        if obj is not None:
            assignment_str = ("".join(self.reference_str_list[:-1])
                              + '.update({"'
                              + self.CurrentHierarchy[-1]
                              + '":obj})')
            if VERBOSITY:
                print(assignment_str)
            eval(assignment_str)
        elif VERBOSITY:
            print("No text?")

        self.CurrentHierarchy.pop()
        self.reference_str_list.pop()
        if tag == "Array":
            self.CurrentHierarchy.pop()
            self.reference_str_list.pop()
        # if necessary advance array index
        if self.CurrentHierarchy:
            ref = self.CurrentHierarchy[-1]
            if isinstance(ref, int):
                ref += 1
                self.CurrentHierarchy[-1] = ref
                self.reference_str_list[
                    len(self.CurrentHierarchy)] = "[" + str(ref) + "]"


##FUNCTIONS

def _attributeconvert(attributes):
    attributestr = ''
    for key in attributes.keys():
        attributestr += ' ' + str(key) + '="' + str(attributes[key]) + '"'
    return attributestr


def get_arts_type(data):
    """Returns the equivalent ARTS data type for a given Python object"""
    # nasty catch for np integers, which aren't behaving as expected
    if isinstance(data, np.integer):
        return 'Index'
    elif MAPPING[type(data)] == 'Array':
        return 'ArrayOf' + get_arts_type(data[0])
    elif MAPPING[type(data)] == 'Tensor':
        return ARTS_TENSOR_NAMES[data.ndim - 1]
    else:
        return MAPPING[type(data)]


def number_to_xml(tag, number, attributes={}):
    return XML_Obj(tag, attributes).write(repr(number) + '\n').finalise().str


def text_to_xml(tag, text, attributes={}):
    return XML_Obj(tag, attributes).write(xml.sax.saxutils.escape(text)
                                          + '\n').finalise().str


def tensor_to_xml(tensor, outfile=None):
    """Store tensor in XML.

    This method takes a Python array  argument and stores it in
    the arts XML format with the appropriate Tag (eg Vector Matrix,
    if outfile (a file object - not a name) is given then it is written to
    that file
    """
    return_string = False
    oldshape = copy.deepcopy(tensor.shape)
    rank = tensor.ndim
    dim_labels = ARTS_DIM_LABELS[:rank]
    dim_labels.reverse()
    if (rank == 1):
        dim_labels[0] = "nelem"
    tag = ARTS_TENSOR_NAMES[rank - 1]
    attributes = {}
    for i in range(rank):
        attributes.update({dim_labels[i]: repr(tensor.shape[i])})
    if (rank > 2):
        # reshape tensor to rank two for output
        tensor = tensor.ravel().reshape(-1, tensor.shape[-1])

    if outfile is None:
        return_string = True
        outfile = XML_Obj(tag, attributes)
        fp = outfile.s
    else:
        outfile.write(
            ('<' + tag + _attributeconvert(attributes) + '>\n').encode())
        fp = outfile

    # use only %.7e, good enough for float
    np.savetxt(fp, tensor, delimiter=" ", fmt="%.7e")
    tensor.reshape(oldshape)
    if return_string:
        return outfile.finalise().str
    else:
        outfile.write(('</' + tag + '>\n').encode())


def save(data, filename):
    """Saves data to arts XML file (filename or stream).

    If the filename ends in '.gz' the XML file will be gzipped
    """
    try:
        data.save(filename)
    except AttributeError:
        if filename.lower().endswith(".gz"):
            opener = contextlib.closing(gzip.GzipFile(filename, "w"))
        else:
            opener = open(filename, "wb")
        with opener as fp:
            outfile = XMLfile(fp)
            outfile.add(data)
            outfile.close()


def load(filename, use_names=True):
    """Loads an ArtsXML-file.

    This general purpose function returns a dictionary structure reflecting
    the structure of the XML file. If there is only one object in the
    structure, then that single object is returned. As far as far as I know,
    this works with every data type exported by ARTS. Note that only in
    special simple cases, such as single tensors, will load return exactly the
    same data given to the save command. The more complicated data structures
    (e.g. SingleScatteringData) have their own load methods. gzipped files can
    be loaded (as long as they end in 'gz').

    Parameters
    ~~~~~~~~~~

    filename : string-like
        Path to file to load (.xml or .xml.gz)

    use_names : boolean, optional (default: True)
        Use name tags in XML-file

    Returns
    ~~~~~~~

    Returns an OrderedDict with the contents of the XML-file.
    """
    if filename.lower().endswith(".gz"):
        opener = contextlib.closing(gzip.GzipFile(filename, "r"))
    else:
        opener = open(filename, "rb")
    handler = Handler(use_names)
    with opener as fp:
        xml.sax.parse(fp, handler)
    data = handler.datastruct["arts"]
    if len(data) == 1:
        return list(data.values())[0]
    else:
        return data



# -*- coding: utf-8 -*-

__version__ = '0.2.1'

from . import arts
from . import files
from . import utils


def _runtest():
    """Run all tests."""
    from os.path import dirname
    import nose
    loader = nose.loader.TestLoader(workingDir=dirname(__file__))
    return nose.run(testLoader=loader)

test = _runtest



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
The arts_types module includes support for various ARTS-classes.

Due to the dynamic nature of Python, some are implemented in a more
generic way. For example, ArrayOf can be easily subclassed to be an np.array
of anything, and all gridded-fields are subclasses of GriddedField.

Classes of special interest may be:

- SingleScatteringData
- ScatteringMetaData

This module allows the generation, manipulation, and input/output in ARTS
XML format of these objects.
"""

from __future__ import print_function

import copy
import gzip
import numbers

import numpy as np

try:
    from cStringIO import StringIO
except:
    from io import StringIO

try:
    unicode = unicode
except NameError:
    # 'unicode' is undefined, must be Python 3
    unicode = str
    basestring = (str, bytes)

from . import artsXML
from . import general
from . import arts_math

PARTICLE_TYPE_GENERAL = 10
PARTICLE_TYPE_MACROS_ISO = 20
PARTICLE_TYPE_HORIZ_AL = 30
PARTICLE_TYPE_SPHERICAL = 40


class BadFieldDimError(general.PyARTSError, ValueError): pass


class ArtsType(object):
    def save(self, f, compressed=None):
        """Writes data to file
        """

        c = self.to_xml()

        if isinstance(f, basestring):
            if compressed is None:
                compressed = f.lower().endswith(".gz")
            if compressed:
                writer = gzip.GzipFile(f, "wb")
            else:
                writer = open(f, "wb")
        else:
            writer = f
        with artsXML.XMLfile(writer, header=True) as out:
            out.write(c)


class ArrayOf(ArtsType, list):
    """Represents an np.array of somethinge

    The type is np.taken from the 'contains' attribute, to be defined by
    subclasses.
    """

    contains = None

    def to_xml(self):
        xml_obj = artsXML.XML_Obj(tag="Array",
                                  attributes={
                                      "type": getattr(self.contains, "name",
                                                      self.contains.__name__),
                                      "nelem": len(self)})
        for elem in self:
            xml_obj.write(elem.to_xml())
        return xml_obj.finalise().str

    def append(self, obj):
        if not isinstance(obj, self.contains):
            raise TypeError("Wrong type in np.array. Expected: %s. Got: %s" %
                            (self.contains, type(obj)))
        super(ArrayOf, self).append(obj)

    def __repr__(self):
        return ("<" + self.__class__.__name__ + "\n["
                + ",\n ".join(repr(elem) for elem in self) + "]" + ">")

    def __getattr__(self, attr):
        L = []
        for elem in self:
            L.append(getattr(elem, attr))
        return L

    @classmethod
    def load(cls, f):
        """Loads ArrayOfSomething

        - f: file to read it from

        Returns ArrayOfSomething
        """
        filedata = artsXML.load(f)
        obj = cls()

        tp = cls.contains
        tpname = cls.contains.__name__
        for artsXML_object in filedata:
            elem = tp._load_from_artsXML_object(artsXML_object[tpname])
            obj.append(elem)
        return obj


_old_ptype_mapping = {
    10: "general",
    20: "macroscopically_isotropic",
    30: "horizontally_aligned",
}


class SingleScatteringData(ArtsType):
    """The class representing the arts SingleScatteringData class.
    
    The data members of this object are identical to the class of the same name in
    ARTS; it includes all the single scattering properties required for
    polarized radiative transfer calculations: the extinction matrix, the
    phase matrix, and the absorption coefficient vector.  The angular,
    frequency, and temperature grids for which these are defined are also
    included.  Another data member - *ptype*, describes the orientational
    symmetry of the particle ensemble, which determines the format of the
    single scattering properties.  The data structure of the ARTS
    SingleScatteringData class is described in the ARTS User
    Guide.

    The methods in the SingleScatteringData class enable the
    calculation of the single scattering properties, and the output of the
    SingleScatteringData structure in the ARTS XML format (see example file).
    The low-level calculations are performed in arts_scat.
    """

    defaults = {"ptype": "macroscopically isotropic",
                # as defined in optproperties.h
                "T_grid": np.array([250]),
                "za_grid": np.arange(0, 181, 10),
                "aa_grid": np.arange(0, 181, 10),
                "equiv_radius": 200,  # equivalent volume radius
                "NP": -1,
                # -1 for spheroid, -2 for cylinder, positive for chebyshev
                # set to None for non-tmatrix (this will make calculations
                # impossible, making this object just a data container)
                'phase': 'ice',
                "aspect_ratio": 1.000001}

    mrr = None
    mri = None

    def __init__(self, params={}, **kwargs):
        """ Constructor

        Parameters
        ----------

        ptype : string
            As for ARTS; see Arts User Guide

        f_grid : 1-D np.array
            np.array for frequency grid [Hz]

        T_grid : 1-D np.array
            np.array for temperature grid [K]

        za_grid : 1-D np.array
            np.array for zenith-angle grid [degree]

        aa_grid : 1-D np.array
            np.array for azimuth-angle grid [degree]

        equiv_radius : number
            equivalent volume radius [micrometer]

        NP : integer or None
            code for shape: -1 for spheroid, -2 for cylinder,
            positive for chebyshev, None for arbitrary shape
            (will not use tmatrix for calculations)

        phase : string
            ice, liquid

        aspect_ratio : number
            Aspect ratio [no unit]

        Some inputs have default values, see SingleScatteringData.defaults.
        """

        # enable keyword arguments
        if kwargs and not params:
            params = kwargs

        params = general.dict_combine_with_default(params,
                                                   self.__class__.defaults)
        # check parameters
        # make sure grids are np np.arrays
        for grid in ['f', 'T', 'za', 'aa']:
            params[grid + '_grid'] = np.array(params[grid + '_grid'])
        if params['aspect_ratio'] == 1:
            raise ValueError(
                "'aspect_ratio' can not be set to exactly 1 due to numerical difficulties in the T-matrix code. use 1.000001 or 0.999999 instead.")

        if "description" in params:
            self.description = params["description"]
        else:
            self.description = "This arts particle file was generated by the arts_scat Python\n" + \
                               "module, which uses the T-matrix code of Mishchenko to calculate single \n" + \
                               "scattering properties. The parameters used to create this file are shown\n" + \
                               "below\n" + str(params)
        for k, v in params.items():
            setattr(self, k, v)

    def to_xml(self):
        xml_obj = artsXML.XML_Obj('SingleScatteringData')
        xml_obj.write(artsXML.text_to_xml("String", self.ptype))
        xml_obj.write(
            artsXML.text_to_xml("String 0",
                                "\"" + general.convert_to_string(
                                    self.description) + "\""))
        xml_obj.write(artsXML.tensor_to_xml(np.array(self.f_grid)))
        xml_obj.write(artsXML.tensor_to_xml(np.array(self.T_grid)))
        xml_obj.write(artsXML.tensor_to_xml(np.array(self.za_grid)))
        xml_obj.write(artsXML.tensor_to_xml(np.array(self.aa_grid)))
        xml_obj.write(artsXML.tensor_to_xml(self.pha_mat_data))
        xml_obj.write(artsXML.tensor_to_xml(self.ext_mat_data))
        xml_obj.write(artsXML.tensor_to_xml(self.abs_vec_data))
        return xml_obj.finalise().str

    @general.force_encoded_string_output
    def __repr__(self):
        S = StringIO()
        S.write("<SingleScatteringData ")
        S.write("ptype={} ".format(self.ptype))
        S.write("phase=%s " % self.phase)
        S.write("equiv_radius=%4e" % self.equiv_radius)
        for nm in ("f_grid", "T_grid", "za_grid", "aa_grid"):
            g = getattr(self, nm)
            S.write(" ")
            if g.size > 1:
                S.write("%s=%4e..%4e" % (nm, g.min(), g.max()))
            elif g.size == 1:
                S.write("%s=%4e" % (nm, float(g.squeeze())))
            else:
                S.write("%s=[]" % nm)
        S.write(">")

        return S.getvalue()

    def integrate(self):
        """Calculate integrated Z over the sphere.

        Only work for azimuthally symmetrically oriented particles.
        """
        if self.ptype == PARTICLE_TYPE_MACROS_ISO:
            return arts_math.integrate_phasemat(
                self.za_grid,
                self.pha_mat_data[..., 0].squeeze())
        else:
            raise general.PyARTSError("Integration implemented only for"
                                      "ptype = %d. Found ptype = %s" %
                                      (PARTICLE_TYPE_MACROS_ISO, self.ptype))

    def normalise(self):
        """Normalises Z to E-A.
        """

        Z_int = self.integrate()
        E_min_A = self.ext_mat_data.squeeze() - self.abs_vec_data.squeeze()
        # use where to prevent divide by zero
        factor = E_min_A / np.where(Z_int == 0, 1, Z_int)
        factor.shape = (factor.size, 1, 1, 1, 1, 1)
        self.pha_mat_data[..., 0] *= factor
        # self.pha_mat_data[..., 0] = self.pha_mat_data[..., 0] * factor

    def normalised(self):
        """Returns normalised copy
        """

        c = copy.deepcopy(self)
        c.normalise()
        return c

    def __getitem__(self, v):
        """Get subset of single-scattering-data

        Must np.take four elements (f, T, za, aa).
        Only implemented for randomly oriented particles.
        """

        if self.ptype != PARTICLE_TYPE_MACROS_ISO:
            raise general.PyARTSError("Slicing implemented only for"
                                      "ptype = %d. Found ptype = %d" %
                                      (PARTICLE_TYPE_MACROS_ISO, self.ptype))
        v2 = list(v)
        for i, el in enumerate(v):
            # to preserve the rank of the data, [n] -> [n:n+1]
            if isinstance(el, numbers.Integral):
                v2[i] = slice(v[i], v[i] + 1, 1)
        f, T, za, aa = v2
        # make a shallow copy (view of the same data)
        c = copy.copy(self)
        c.f_grid = c.f_grid[f]
        c.T_grid = c.T_grid[T]
        c.za_grid = c.za_grid[za]
        c.aa_grid = c.aa_grid[aa]
        c.ext_mat_data = c.ext_mat_data[f, T, :, :, :]
        c.pha_mat_data = c.pha_mat_data[f, T, za, aa, :, :, :]
        c.abs_vec_data = c.abs_vec_data[f, T, :, :, :]
        c.checksize()
        return c

    def checksize(self):
        """Verifies size is consistent.

        raises PyARTSError if not. Otherwise, do nothing.
        """
        if not ((self.f_grid.size or 1, self.T_grid.size or 1) ==
                    self.ext_mat_data.shape[:2] ==
                    self.pha_mat_data.shape[:2] ==
                    self.abs_vec_data.shape[:2] and
                        (self.za_grid.size or 1, self.aa_grid.size or 1) ==
                        self.pha_mat_data.shape[2:4]):
            raise general.PyARTSError(
                "Inconsistent sizes in SingleScatteringData.\n"
                "f_grid: %s, T_grid: %s, za_grid: %s, aa_grid: %s, "
                "ext_mat: %s, pha_mat: %s, abs_vec: %s" %
                (self.f_grid.size or 1, self.T_grid.size or 1,
                 self.za_grid.size or 1, self.aa_grid.size or 1,
                 self.ext_mat_data.shape, self.pha_mat_data.shape,
                 self.abs_vec_data.shape))

    def assp2backcoef(self):
        """The function returns the radar backscattering coeffcient. This is the
        phase function times 4pi, following the standard definition in the radar
        community.

        Returns:
            Backscattering coefficients, one value for each frequency and
            temperature in S. [m2]
        """

        back_coef = np.multiply(4*np.pi, self.pha_mat_data[:, :, -1, 0, 0, 0,
                                                           0])
        return back_coef

    def assp2g(self):
        """For a normalised phase function (p), g equals the 4pi integral of
        p*cos(th), where th is the scattering angle. For pure isotropic
        scattering g = 0, while pure forward scattering has g=1.

        Warning, this function does not handle the extreme cases of
        delta-function type of forward or backward scattering lobes. A g of
        zero is returned for these cases.

        Returns:
            Backscattering coefficients, one value for each frequency and
            temperature in S. [m2]
        """

        g = np.zeros((len(self.f_grid), len(self.T_grid)))
        # ARTS uses pure phase matrix values, and not a normalised phase
        # function, and we need to include a normalisation.

        za_rad_grid = np.radians([self.za_grid])

        aziWeight = abs(np.sin(za_rad_grid))
        cosTerm = np.cos(za_rad_grid)

        for j in range(0, len(self.f_grid)):
            for i in range(0, len(self.T_grid)):
                phase_grid = self.pha_mat_data[j, i, :, 0, 0, 0, 0]

                normFac = np.trapz(np.multiply(
                    phase_grid, aziWeight), za_rad_grid)

                if normFac == 0:
                # If normFac is zero, this means that phase_grid==0 and
                # should indicate very small particles that have g=0.
                    g[j, i] = 0
                else:
                    temp_cosPhase = np.multiply(cosTerm, phase_grid)

                    temp = np.trapz(np.multiply(temp_cosPhase, aziWeight),
                                    za_rad_grid)

                    g[j, i] = np.divide(temp, normFac)
        return g

    def checkassp(self):
        """Verfies properties of SSP.

        Raises:
            PyARTSError: If ptype is not macroscopically isotropic, or if first
                and last value of za_grid does not equal exactly 0 and 180
                respectively.
        """

        if self.ptype != "macroscopically_isotropic":
            raise general.PyARTSError("So far just complete random"
                                      " orientation is handled.")
        if self.za_grid[0] != 0:
            raise general.PyARTSError("First value of za_grid must be 0.")
        if self.za_grid[-1] != 180:
            raise general.PyARTSError("Last value of za_grid must be 180.")

    @classmethod
    def _load_from_artsXML_object(cls, artsXML_object):
        """Loads a SingleScatteringData object from an artsXML_object.
        """

        params = {
            "f_grid": artsXML_object['Vector'],
            "T_grid": artsXML_object['Vector 0'],
            "za_grid": artsXML_object['Vector 1'],
            "aa_grid": artsXML_object['Vector 2'],
        }

        obj = cls(**params)
        if 'String 0' in artsXML_object:
            obj.description = artsXML_object['String 0'][1:-1],
            obj.ptype = artsXML_object['String'][1:-1]
            if obj.ptype not in _old_ptype_mapping.values():
                raise RuntimeError("Illegal ptype: {}".format(obj.ptype))
        else:
            obj.description = artsXML_object['String'][1:-1],
            obj.ptype = _old_ptype_mapping[artsXML_object['Index']]

        obj.pha_mat_data = artsXML_object['Tensor7']
        obj.ext_mat_data = artsXML_object['Tensor5']
        obj.abs_vec_data = artsXML_object['Tensor5 0']
        return obj

    @classmethod
    def load(cls, filename):
        """Loads a SingleScatteringData object from an existing file.

        Note that this can only import data members that are actually in the file
        - so the scattering properties may not be consistent with the
        *params* data member.
        """
        scat_data = artsXML.load(filename)
        return cls._load_from_artsXML_object(scat_data)


class ScatteringMetaData(ArtsType):
    """Represents a ScatteringMetaData object.

    See online ARTS documentation for object details.
    """

    def __init__(self, description, source, refr_index, mass, diameter_max,
                 diameter_volume_equ, diameter_area_equ_aerodynamical):
        self.description = description
        self.source = source
        self.refr_index = refr_index
        self.mass = mass
        self.diameter_max = diameter_max
        self.diameter_volume_equ = diameter_volume_equ
        self.diameter_area_equ_aerodynamical = diameter_area_equ_aerodynamical

    @general.force_encoded_string_output
    def __repr__(self):
        S = StringIO()
        S.write("<ScatteringMetaData ")
        S.write("description=%s " % self.description)
        S.write("source=%s " % self.source)
        S.write("refr_index=%s " % self.refr_index)
        S.write("mass=%4e " % self.mass)
        S.write("diameter_max=%4e " % self.diameter_max)
        S.write("diameter_volume_equ=%4e " % self.diameter_volume_equ)
        S.write(
            "diameter_area_equ_aerodynamical=%4e " % self.diameter_area_equ_aerodynamical)
        return S.getvalue()

    @classmethod
    def _load_from_artsXML_object(cls, artsXML_object):
        """Loads a SingleScatteringData object from an artsXML_object.
        """

        params = {
            "description": artsXML_object['String'][1:-1],
            "source": artsXML_object['String 0'][1:-1],
            "refr_index": artsXML_object['String 1'][1:-1],
            "mass": artsXML_object['Numeric'],
            "diameter_max": artsXML_object['Numeric 0'],
            "diameter_volume_equ": artsXML_object['Numeric 1'],
            "diameter_area_equ_aerodynamical": artsXML_object['Numeric 2'],
        }

        obj = cls(**params)
        return obj

    @classmethod
    def load(cls, filename):
        """Loads a SingleScatteringData object from an existing file.

        Note that this can only import data members that are actually in the file
        - so the scattering properties may not be consistent with the
        *params* data member.
        """
        scat_data = artsXML.load(filename)
        return cls._load_from_artsXML_object(scat_data)

    def to_xml(self):
        xml_obj = artsXML.XML_Obj('ScatteringMetaData',
                                  attributes={'version': 3})

        xml_obj.write(
            artsXML.text_to_xml("String",
                                "\"" + general.convert_to_string(
                                    self.description) + "\""))
        xml_obj.write(
            artsXML.text_to_xml("String",
                                "\"" + general.convert_to_string(
                                    self.source) + "\""))
        xml_obj.write(
            artsXML.text_to_xml("String",
                                "\"" + general.convert_to_string(
                                    self.refr_index) + "\""))
        xml_obj.write(artsXML.number_to_xml("Numeric", self.mass))
        xml_obj.write(artsXML.number_to_xml("Numeric", self.diameter_max))
        xml_obj.write(
            artsXML.number_to_xml("Numeric", self.diameter_volume_equ))
        xml_obj.write(artsXML.number_to_xml("Numeric",
                                            self.diameter_area_equ_aerodynamical))

        return xml_obj.finalise().str


class ArrayOfSingleScatteringData(ArrayOf):
    """Represents an ArrayOfSingleScatteringData.
    """

    contains = SingleScatteringData

    def normalise(self):
        """Call .normalise() on each SSD
        """

        for elem in self:
            elem.normalise()

    def normalised(self):
        """Return normalised copy of self
        """

        c = copy.deepcopy(self)
        for elem in c:
            elem.normalise()
        return c

    def integrate(self):
        """Return list with all integrated values.
        """

        L = []
        for elem in self:
            L.append(elem.integrate())
        return L

    @property
    def sub(self):
        """Get subslice of each element.
        """

        class Subber(object):
            def __getitem__(self2, *args):
                L = self.__class__()
                for elem in self:
                    L.append(elem.__getitem__(*args))
                return L

        return Subber()

    @classmethod
    def fromData(cls, name, shapestr, sizes, frequencies, angles, data):
        """Load from data.

        Do not call directly, call fromYang, fromHong, etc.
        """

        obj = cls()
        obj.smd = ArrayOfScatteringMetaData()
        for (i, size) in enumerate(sizes):
            # add SSD/SMD, one by one
            obj.append(
                cls.dispatcher_SSD[name](size, shapestr, frequencies,
                                         angles,
                                         data[:, i]))
            obj.smd.append(
                cls.dispatcher_SMD[name](size, shapestr, frequencies,
                                         angles,
                                         data[:, i]))
        return obj


class ArrayOfScatteringMetaData(ArrayOf):
    contains = ScatteringMetaData



"""
This module includes general purpose math functions.  This module was developed before
scipy was included as a prerequisite, so there will be some functions remaining that
duplicate scipy functionality. 
"""
import numpy as np
import scipy as sp


def integrate_phasemat(angles, Z):
    """Integrates the phase-matrix over all angles

    Parameters
    ~~~~~~~~~~

    angles : 1D-array
        Angles to integrate over [degrees]

    Z : nD-array
        quantity to integrate. The last dimension must match the size of
        angles.

    Returns
    ~~~~~~~

    Z_int : (n-1)D-array
        Array with integrated Z. Has all but the last dimension of Z.
    """

    angles = np.deg2rad(angles)
    integrand = 2 * np.pi * Z * np.sin(angles)
    return sp.integrate.trapz(integrand, angles)
